import unittest
from psyki.logic.datalog.grammar.adapters.antlr4 import get_formula_from_string


class TestFormula(unittest.TestCase):
    virginica_rule = "class(PL,PW,SL,SW,virginica) ← PL > 2.28 ∧ PW > 1.64"
    expected_formula_structure = "class(PL,PW,SL,SW,virginica)←((((PL)>(2.28)))∧(((PW)>(1.64))))"

    def test_parsing_with_antlr4(self):
        formula = get_formula_from_string(self.virginica_rule)
        self.assertEqual(self.expected_formula_structure, str(formula))



from os.path import isdir
from psyki.resources import PATH, create_antlr4_parser
if not isdir(str(PATH / 'dist')):
    create_antlr4_parser(str(PATH / 'Datalog.g4'), str(PATH / 'dist'))
from antlr4 import CommonTokenStream, InputStream
from psyki.logic.datalog import *
from psyki.resources.dist.DatalogLexer import DatalogLexer
from psyki.resources.dist.DatalogParser import DatalogParser


def get_formula_from_string(rule: str) -> DatalogFormula:
    return get_formula(DatalogParser(CommonTokenStream(DatalogLexer(InputStream(rule)))).formula())


def get_formula(ast: DatalogParser.FormulaContext) -> DatalogFormula:
    return DatalogFormula(_get_definition_clause(ast.lhs), _get_clause(ast.rhs), ast.op.text)


def _get_definition_clause(node: DatalogParser.DefPredicateArgsContext):
    return DefinitionClause(node.pred.text, _get_arguments(node.args))


def _get_arguments(node: DatalogParser.MoreArgsContext or DatalogParser.LastTermContext):
    if isinstance(node, DatalogParser.MoreArgsContext):
        return Argument(_get_term(node.name), _get_arguments(node.args))
    elif isinstance(node, DatalogParser.LastTermContext):
        return Argument(_get_term(node.name))


def _get_clause(node: DatalogParser.ClauseExpressionContext or DatalogParser.ClauseExpressionNoParContext
                      or DatalogParser.ClauseLiteralContext or DatalogParser.ClauseClauseContext):
    if isinstance(node, DatalogParser.ClauseExpressionContext)\
            or isinstance(node, DatalogParser.ClauseExpressionNoParContext):
        return Expression(_get_clause(node.left), _get_clause(node.right), node.op.text)
    elif isinstance(node, DatalogParser.ClauseLiteralContext):
        return _get_literal(node.lit)
    elif isinstance(node, DatalogParser.ClauseClauseContext):
        return _get_clause(node.c)


def _get_literal(node: DatalogParser.LiteralPredContext or DatalogParser.LiteralNegContext):
    if isinstance(node, DatalogParser.LiteralNegContext):
        return Negation(_get_clause(node.pred))
    elif isinstance(node, DatalogParser.LiteralPredContext):
        return _get_predicate(node.pred)


def _get_predicate(node: DatalogParser.PredicateTermContext or DatalogParser.PredicateUnaryContext
                         or DatalogParser.PredicateArgsContext):
    if isinstance(node, DatalogParser.PredicateTermContext):
        return _get_term(node.name)
    elif isinstance(node, DatalogParser.PredicateUnaryContext):
        return Unary(node.pred.text)
    elif isinstance(node, DatalogParser.PredicateArgsContext):
        return Nary(node.pred.text, _get_arguments(node.args))


def _get_term(node: DatalogParser.TermVarContext or DatalogParser.TermConstContext):
    if isinstance(node, DatalogParser.TermVarContext):
        return Variable(node.var.text)
    elif isinstance(node, DatalogParser.TermConstContext):
        return _get_constant(node.name)


def _get_constant(node: DatalogParser.ConstNumberContext or DatalogParser.ConstNameContext):
    if isinstance(node, DatalogParser.ConstNumberContext):
        return Number(node.num.text)
    elif isinstance(node, DatalogParser.ConstNameContext):
        return Predication(node.name.text)



import unittest
from psyki.logic.datalog import SubNetworkBuilder
from psyki.logic.datalog.grammar.adapters.antlr4 import get_formula_from_string
from tensorflow.keras import Input, Model
from test.resources.data import data_to_int, get_binary_data, get_dataset_dataframe
from test.resources.rules import get_rules
from test.resources.data.splice_junction import CLASS_MAPPING, AGGREGATE_FEATURE_MAPPING, \
    get_splice_junction_extended_feature_mapping
from test.resources.rules.splice_junction import get_splice_junction_datalog_rules, get_binary_datalog_rules


class TestSubnetworkBuilder(unittest.TestCase):

    rules = get_rules('splice_junction')
    rules = get_splice_junction_datalog_rules(rules)
    rules = get_binary_datalog_rules(rules)
    data = get_dataset_dataframe('splice_junction')
    y = data_to_int(data.iloc[:, -1:], CLASS_MAPPING)
    x = get_binary_data(data.iloc[:, :-1], AGGREGATE_FEATURE_MAPPING)
    y.columns = [x.shape[1]]
    data = x.join(y)
    rules = [get_formula_from_string(rule) for rule in rules]
    inputs = Input((240,))
    fuzzifier = SubNetworkBuilder(inputs, get_splice_junction_extended_feature_mapping())
    modules = fuzzifier.visit(rules)

    def test_rule_correctness(self):
        predict_ie = Model(self.inputs, self.modules[1])
        result_ie = predict_ie.predict(self.data.iloc[:, :-1]).astype(bool)[:, -1]
        predict_ei = Model(self.inputs, self.modules[0])
        result_ei = predict_ei.predict(self.data.iloc[:, :-1]).astype(bool)[:, -1]
        result_n = (~ result_ei) & (~ result_ie)

        self.assertTrue(sum(result_ie & (self.data.iloc[:, -1] == CLASS_MAPPING['ie'])), 295)
        self.assertTrue(sum(result_ei & (self.data.iloc[:, -1] == CLASS_MAPPING['ei'])), 31)
        self.assertTrue(sum(result_n & (self.data.iloc[:, -1] == CLASS_MAPPING['n'])), 1652)


if __name__ == '__main__':
    unittest.main()



import csv
from pathlib import Path
from test.resources.rules.iris import PATH as IRIS_PATH
from test.resources.rules.poker import PATH as POKER_PATH
from test.resources.rules.splice_junction import PATH as SJ_PATH


PATH = Path(__file__).parents[0]

RULES_REGISTER = {
    "iris": IRIS_PATH,
    "poker": POKER_PATH,
    "splice_junction": SJ_PATH
}


def get_rules(rule_domain: str = "iris", rule_name: str = "kb") -> list[str]:
    result = []
    with open(str(RULES_REGISTER[rule_domain] / rule_name) + '.txt', mode="r") as file:
        reader = csv.reader(file, delimiter=';')
        for item in reader:
            result += item
    return result



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Callable, List
from tensorflow import maximum, minimum, constant, SparseTensor, cast, reshape, reduce_max, tile, Tensor, gather
from tensorflow.keras.backend import to_dense
from tensorflow.keras.layers import Minimum, Maximum, Dense, Dot, Concatenate
from tensorflow.keras.layers import Lambda
from tensorflow.python.ops.array_ops import shape
from tensorflow.python.ops.init_ops import constant_initializer, Ones, Zeros
from psyki.logic.datalog.grammar import DatalogFormula, Expression, Variable, Number, Unary, Predication, \
    DefinitionClause, Argument, Nary, Negation, Clause
from psyki.ski import Fuzzifier, Formula
from psyki.utils import eta, eta_one_abs, eta_abs_one


class DatalogFuzzifier(Fuzzifier, ABC):
    feature_mapping: dict[str, int] = {}
    classes = {}
    _predicates: dict[str, tuple[dict[str, int], Callable]] = {}

    def __init__(self):
        self.visit_mapping: dict[Formula.__class__, Callable] = {
            DatalogFormula: self._visit_formula,
            Expression: self._visit_expression,
            Negation: self._visit_negation,
            Variable: self._visit_variable,
            Number: self._visit_number,
            Unary: self._visit_unary,
            Nary: self._visit_nary
        }

    def visit(self, rules: List[Formula]) -> Any:
        self._clear()

    @abstractmethod
    def _visit(self, formula: Formula, local_mapping: dict[str, int] = None) -> Any:
        pass

    @abstractmethod
    def _visit_formula(self, formula: Formula, local_mapping: dict[str, int] = None) -> Any:
        pass

    @abstractmethod
    def _visit_expression(self, formula: Formula, local_mapping: dict[str, int] = None) -> Any:
        pass

    @abstractmethod
    def _visit_negation(self, formula: Formula, local_mapping: dict[str, int] = None) -> Any:
        pass

    @abstractmethod
    def _visit_variable(self, formula: Formula, local_mapping: dict[str, int] = None) -> Any:
        pass

    @abstractmethod
    def _visit_number(self, formula: Formula, _) -> Any:
        pass

    @abstractmethod
    def _visit_unary(self, formula: Formula, _) -> Any:
        pass

    def _visit_nary(self, node: Nary, local_mapping: dict[str, int] = None):
        # Prevents side effect on the original local map.
        local_mapping_copy = self._predicates[node.name][0].copy()
        inv_map = {v: k for k, v in local_mapping_copy.items()}
        # Dynamic bounding between the variables of the caller and the callee.
        for i, variable in enumerate(self._get_variables_names(node.arg)):
            if i in inv_map.keys():
                if variable in self.feature_mapping:
                    local_mapping_copy[inv_map[i]] = self.feature_mapping.get(variable)
                elif variable in local_mapping:
                    local_mapping_copy[inv_map[i]] = local_mapping.get(variable)
        return self._predicates[node.name][1](local_mapping_copy)

    def _get_variables_names(self, node: Argument) -> list[str]:
        if node is not None and isinstance(node.term, Variable):
            return [node.term.name] + self._get_variables_names(node.arg)
        else:
            return []

    def _get_predication_name(self, node: Argument):
        if node is not None and node.arg is not None:
            return self._get_predication_name(node.arg)
        elif node is not None and isinstance(node.term, Predication):
            return node.term.name
        else:
            return None

    @abstractmethod
    def _clear(self):
        pass


class ConstrainingFuzzifier(DatalogFuzzifier, ABC):
    """
    A fuzzifier that encodes logic formulae into continuous functions (or something equivalent) to constrain the
    behaviour of the predictor during the training in such a way that it is penalised when it violates the prior
    knowledge.
    """
    def visit(self, rules: List[Formula]) -> Any:
        super().visit(rules)
        for rule in rules:
            self._visit(rule, {})
        return self.classes


class StructuringFuzzifier(DatalogFuzzifier, ABC):
    """
    A fuzzifier that encodes logic formulae into new sub parts of the predictors which mimic the logic formulae.
    """
    def visit(self, rules: List[Formula]) -> Any:
        super().visit(rules)
        for rule in rules:
            self._visit(rule, {})
        return list(self.classes.values())


class Lukasiewicz(ConstrainingFuzzifier):
    """
    Fuzzifier that implements a mapping from crispy logic rules into a continuous interpretation inspired by the
    mapping of Lukasiewicz. The resulting object is a list of continuous functions that can be used to constraint
    the predictor during its training. This is suitable for classification tasks.
    """

    def __init__(self, class_mapping: dict[str, int], feature_mapping: dict[str, int]):
        """
        @param class_mapping: a map between constants representing the expected class in the logic formulae and the
        corresponding index for the predictor. Example:
            - 'setosa': 0,
            - 'virginica': 1,
            - 'versicolor': 2.
        @param feature_mapping: a map between variables in the logic formulae and indices of dataset features. Example:
            - 'PL': 0,
            - 'PW': 1,
            - 'SL': 2,
            - 'SW': 3.
        """
        super().__init__()
        self.feature_mapping = feature_mapping
        self.class_mapping = {string: cast(to_dense(SparseTensor([[0, index]], [1.], (1, len(class_mapping)))), float)
                              for string, index in class_mapping.items()}
        self.classes: dict[str, Callable] = {}
        self._rhs: dict[str, Callable] = {}
        self._operation = {
            '∧': lambda l, r: lambda x: eta(maximum(l(x), r(x))),
            '∨': lambda l, r: lambda x: eta(minimum(l(x), r(x))),
            '→': lambda l, r: lambda x: eta(l(x) - r(x)),
            '↔': lambda l, r: lambda x: eta(abs(l(x) - r(x))),
            '=': lambda l, r: lambda x: eta(abs(l(x) - r(x))),
            '<': lambda l, r: lambda x: eta(constant(1.) - eta(minimum(eta(constant(1.) - maximum(l(x) - r(x))),
                                                                       eta(abs(l(x) - r(x)))))),
            '≤': lambda l, r: lambda x: eta(constant(1.) - eta(constant(1.) - maximum(constant(0.), l(x) - r(x)))),
            '>': lambda l, r: lambda x: eta(constant(1.) - maximum(constant(0.), l(x) - r(x))),
            '≥': lambda l, r: lambda x: eta(minimum(eta(constant(1.) - maximum(l(x) - r(x))), eta(abs(l(x) - r(x))))),
            'm': lambda l, r: lambda x: minimum(l(x), r(x)),
            '+': lambda l, r: lambda x: l(x) + r(x),
            '*': lambda l, r: lambda x: l(x) * r(x)
        }

    def _clear(self):
        self.classes = {}
        self._rhs = {}
        self._predicates = {}

    def _visit(self, formula: Formula, local_mapping: dict[str, int] = None) -> Callable:
        return self.visit_mapping.get(formula.__class__)(formula, local_mapping)

    def _visit_formula(self, node: DatalogFormula, local_mapping: dict[str, int] = None) -> None:
        self._visit_definition_clause(node.lhs, node.rhs, local_mapping)

    def _visit_definition_clause(self, node: DefinitionClause, rhs: Clause,
                                 local_mapping: dict[str, int] = None) -> None:
        definition_name = node.predication
        predication_name = self._get_predication_name(node.arg)

        if predication_name is not None:
            class_tensor = reshape(self.class_mapping[predication_name], (1, len(self.class_mapping)))
            l = lambda y: eta(reduce_max(abs(tile(class_tensor, (shape(y)[0], 1)) - y), axis=1))
            r = self._visit(rhs, local_mapping)
            if predication_name not in self.classes.keys():
                self.classes[predication_name] = lambda x, y: eta(r(x) - l(y))
                self._rhs[predication_name] = lambda x: r(x)
            else:
                incomplete_function = self._rhs[predication_name]
                self.classes[predication_name] = lambda x, y: eta(minimum(incomplete_function(x), r(x)) - l(y))
                self._rhs[predication_name] = lambda x: minimum(incomplete_function(x), r(x))
        else:
            # Substitute variables that are not matching features with mapping functions
            variables_names = self._get_variables_names(node.arg)
            for i, variable in enumerate(variables_names):
                if variable not in self.feature_mapping.keys():
                    local_mapping[variable] = i

            if definition_name not in self._predicates.keys():
                self._predicates[definition_name] = (local_mapping, lambda m: lambda x: self._visit(rhs, m)(x))
            else:
                incomplete_function = self._predicates[definition_name][1]
                self._predicates[definition_name] = (local_mapping,
                                                     lambda m: lambda x: eta(minimum(incomplete_function(m)(x),
                                                                                     self._visit(rhs, m)(x))))

    def _visit_expression(self, node: Expression, local_mapping: dict[str, int] = None) -> Callable:
        l, r = self._visit(node.lhs, local_mapping), self._visit(node.rhs, local_mapping)
        return self._operation.get(node.op)(l, r)

    def _visit_variable(self, node: Variable, local_mapping: dict[str, int] = None):
        if node.name in self.feature_mapping.keys():
            return lambda x: x[:, self.feature_mapping[node.name]]
        elif node.name in local_mapping.keys():
            return lambda x: x[:, local_mapping[node.name]]
        else:
            raise Exception("No match between variable name and feature names.")

    def _visit_number(self, node: Number, _):
        return lambda _: node.value

    def _visit_unary(self, node: Unary, _):
        return self._predicates[node.name][1]({})

    def _visit_negation(self, node: Negation, local_mapping: dict[str, int] = None):
        return lambda x: eta(constant(1.) - self._visit(node.predicate, local_mapping)(x))


class SubNetworkBuilder(StructuringFuzzifier):
    """
    Fuzzifier that implements a mapping from crispy logic rules into small neural networks which mimic the prior
    knowledge with a continuous interpretation. The resulting object is a list of modules (ad hoc layers) that can be
    exploited by the predictor during and after its training. This is suitable for classification and regression tasks.
    """

    def __init__(self, predictor_input: Tensor, feature_mapping: dict[str, int]):
        """
        @param predictor_input: the input tensor of the predictor.
        @param feature_mapping: a map between variables in the logic formulae and indices of dataset features. Example:
            - 'PL': 0,
            - 'PW': 1,
            - 'SL': 2,
            - 'SW': 3.
        """
        super().__init__()
        self.predictor_input = predictor_input
        self.feature_mapping = feature_mapping
        self.classes: dict[str, Tensor] = {}
        self.__rhs: dict[str, Tensor] = {}
        self._trainable = False
        self._operation = {
            '→': lambda l: None,
            '↔': lambda l: None,
            '∧': lambda l: Minimum()(l),
            '∨': lambda l: Maximum()(l),
            '+': lambda l: Dense(1, kernel_initializer=Ones, activation='linear', trainable=self._trainable)
            (Concatenate(axis=1)(l)),
            '=': lambda l: Dense(1, kernel_initializer=constant_initializer([1, -1]),
                                 activation=eta_one_abs, trainable=self._trainable)(Concatenate(axis=1)(l)),
            '<': lambda l: Dense(1, kernel_initializer=constant_initializer([-1, 1]),
                                 activation=eta, trainable=self._trainable)(Concatenate(axis=1)(l)),
            '≤': lambda l: Maximum()([Dense(1, kernel_initializer=constant_initializer([-1, 1]),
                                            activation=eta, trainable=self._trainable)(Concatenate(axis=1)(l)),
                                      Dense(1, kernel_initializer=constant_initializer([1, -1]), activation=eta_one_abs,
                                            trainable=self._trainable)(Concatenate(axis=1)(l))]),
            '>': lambda l: Dense(1, kernel_initializer=constant_initializer([1, -1]),
                                 activation=eta, trainable=self._trainable)(Concatenate(axis=1)(l)),
            '≥': lambda l: Maximum()([Dense(1, kernel_initializer=constant_initializer([1, -1]),
                                            activation=eta, trainable=self._trainable)(Concatenate(axis=1)(l)),
                                      Dense(1, kernel_initializer=constant_initializer([1, -1]), activation=eta_one_abs,
                                            trainable=self._trainable)(Concatenate(axis=1)(l))]),
            'm': lambda l: Minimum()(l),
            '*': lambda l: Dot(axes=1)(l)
        }

    def _clear(self):
        self.classes = {}
        self.__rhs = {}
        self._predicates = {}
        self._trainable = False

    def _visit(self, formula: Formula, local_mapping: dict[str, int] = None) -> Any:
        return self.visit_mapping.get(formula.__class__)(formula, local_mapping)

    def _visit_formula(self, node: DatalogFormula, local_mapping: dict[str, int] = None):
        # if the implication symbol is a double left arrow '⇐', then the weights of the module are trainable.
        self._trainable = node.op == '⇐'
        self._visit_definition_clause(node.lhs, node.rhs, local_mapping)

    def _visit_definition_clause(self, node: DefinitionClause, rhs: Clause, local_mapping: dict[str, int] = None):
        definition_name = node.predication
        predication_name = self._get_predication_name(node.arg)

        if predication_name is not None:
            if predication_name not in self.classes.keys():
                self.classes[predication_name] = self._visit(rhs, local_mapping)
                self.__rhs[predication_name] = self._visit(rhs, local_mapping)
            else:
                incomplete_function = self.__rhs[predication_name]
                self.classes[predication_name] = maximum(incomplete_function, self._visit(rhs, local_mapping))
                self.__rhs[predication_name] = maximum(incomplete_function, self._visit(rhs, local_mapping))
        else:
            # Substitute variables that are not matching features with mapping functions
            variables_names = self._get_variables_names(node.arg)
            for i, variable in enumerate(variables_names):
                if variable not in self.feature_mapping.keys():
                    local_mapping[variable] = i

            if definition_name not in self._predicates.keys():
                self._predicates[definition_name] = (local_mapping, lambda m: self._visit(rhs, m))
            else:
                incomplete_function = self._predicates[definition_name][1]
                self._predicates[definition_name] = (local_mapping,
                                                     lambda m: maximum(incomplete_function(m), self._visit(rhs, m)))

    def _visit_expression(self, node: Expression, local_mapping: dict[str, int] = None):
        if len(node.nary) < 1:
            previous_layer = [self._visit(node.lhs, local_mapping), self._visit(node.rhs, local_mapping)]
        else:
            previous_layer = [self._visit(clause, local_mapping) for clause in node.nary]
        return self._operation.get(node.op)(previous_layer)

    def _visit_variable(self, node: Variable, local_mapping: dict[str, int] = None):
        if node.name in self.feature_mapping.keys():
            return Lambda(lambda x: gather(x, [self.feature_mapping[node.name]], axis=1))(self.predictor_input)
        elif node.name in local_mapping.keys():
            return Lambda(lambda x: gather(x, [local_mapping[node.name]], axis=1))(self.predictor_input)
        else:
            raise Exception("No match between variable name and feature names.")

    def _visit_number(self, node: Number, _):
        return Dense(1, kernel_initializer=Zeros,
                     bias_initializer=constant_initializer(node.value),
                     trainable=False, activation='linear')(self.predictor_input)

    def _visit_unary(self, node: Unary, _):
        return self._predicates[node.name][1]({})

    def _visit_negation(self, node: Negation, local_mapping: dict[str, int] = None):
        return Dense(1, kernel_initializer=Ones, activation=eta_abs_one, trainable=False) \
            (self._visit(node.predicate, local_mapping))



from pathlib import Path

PATH = Path(__file__).parents[0]

FEATURE_MAPPING = {
        'S1': 0,
        'R1': 1,
        'S2': 2,
        'R2': 3,
        'S3': 4,
        'R3': 5,
        'S4': 6,
        'R4': 7,
        'S5': 8,
        'R5': 9
    }

CLASS_MAPPING = {
        'nothing': 0,
        'pair': 1,
        'two': 2,
        'three': 3,
        'straight': 4,
        'flush': 5,
        'full': 6,
        'four': 7,
        'straight_flush': 8,
        'royal_flush': 9
    }


from pathlib import Path

PATH = Path(__file__).parents[0]


def get_indices() -> list[int]:
    return list(range(-30, 0)) + list(range(1, 31))


FEATURES = ['a', 'c', 'g', 't']

CLASS_MAPPING = {'ei': 0,
                 'ie': 1,
                 'n': 2}

AGGREGATE_FEATURE_MAPPING = {'a': ('a',),
                             'c': ('c',),
                             'g': ('g',),
                             't': ('t',),
                             'd': ('a', 'g', 't'),
                             'm': ('a', 'c'),
                             'n': ('a', 'c', 'g', 't'),
                             'r': ('a', 'g'),
                             's': ('c', 'g'),
                             'y': ('c', 't')}

SPLICE_JUNCTION_INDICES = get_indices()


def get_splice_junction_feature_mapping() -> dict[str: int]:
    return _get_feature_mapping(SPLICE_JUNCTION_INDICES)


def get_splice_junction_extended_feature_mapping() -> dict[str: int]:
    return _get_extended_feature_mapping(FEATURES, SPLICE_JUNCTION_INDICES)


def _get_feature_mapping(variable_indices: list[int]) -> dict[str: int]:
    return {'X' + ('_' if j < 0 else '') + str(abs(j)): i for i, j in enumerate(variable_indices)}


def _get_extended_feature_mapping(features: list[str], variable_indices: list[int]) -> dict[str: int]:
    result = {'X' + ('_' if j < 0 else '') + str(abs(j)) + f: k + i * len(features)
            for i, j in enumerate(variable_indices) for k, f in enumerate(features)}
    return result



import re
from pathlib import Path
from typing import Any, Iterable
import numpy as np
import pandas as pd
from test.resources.data.poker import PATH as POKER_PATH
from test.resources.data.splice_junction import PATH as SJ_PATH


PATH = Path(__file__).parents[0]
DATA_REGISTER = {
    "poker": POKER_PATH,
    "splice_junction": SJ_PATH
}


def get_dataset(dataset_domain: str = "poker", dataset_name: str = "data") -> Any:
    return np.genfromtxt(str(DATA_REGISTER[dataset_domain] / dataset_name) + '.csv', delimiter=',', dtype=float)


def get_dataset_dataframe(dataset_domain: str = "poker", dataset_name: str = "data") -> pd.DataFrame:
    return _get_data(str(DATA_REGISTER[dataset_domain] / dataset_name) + '.txt')


def data_to_int(data: pd.DataFrame, mapping: dict[str: int]) -> pd.DataFrame:
    return data.applymap(lambda x: mapping[x] if x in mapping.keys() else x)


def _get_data(file: str) -> pd.DataFrame:
    x = []
    with open(file) as f:
        for row in f:
            row = re.sub('\n', '', row)
            label, _, features = row.split(',')
            features = list(f for f in features.lower())
            features.append(label.lower())
            x.append(features)
    return pd.DataFrame(x)


def get_binary_data(data: pd.DataFrame, mapping: dict[str: set[str]]) -> pd.DataFrame:
    sub_features = sorted(_get_values(mapping))
    results = []
    for _, row in data.iterrows():
        row_result = []
        for value in row:
            positive_features = mapping[value]
            for feature in sub_features:
                row_result.append(1 if feature in positive_features else 0)
        results.append(row_result)
    return pd.DataFrame(results, dtype=int)


def _get_values(mapping: dict[str: set[str]]) -> Iterable[str]:
    result = set()
    for values_set in mapping.values():
        for value in values_set:
            result.add(value)
    return result



from typing import List, Any
import numpy as np
import tuprolog.solve
import tuprolog.theory.parsing
from tuprolog.solve.prolog import prolog_solver
from psyki.logic.prolog.grammar import PrologFormula
from psyki.ski import Fuzzifier, Formula


class EnricherFuzzifier(Fuzzifier):

    def __init__(self, knowledge_base_file: str, mapping=None):
        self.engine = prolog_solver(static_kb=EnricherFuzzifier._parse_file(knowledge_base_file))
        self.mapping = mapping

    def visit(self, rules: List[Formula]) -> Any:
        if all(isinstance(rule, PrologFormula) for rule in rules):
            substitutions = [self.engine.solveOnce(query.theory) for index, query in enumerate(rules)]
            results = [str(query.solved_query.get_arg_at(1)) if query.is_yes else -1 for query in substitutions]
            processed_results = [float(result) if result.isnumeric() else self.mapping[result] for result in results]
            return np.array(processed_results)
        else:
            raise Exception('Trying to visit a not Prolog Formula in a Prolog Fuzzifier')

    @staticmethod
    def _parse_file(file: str):
        knowledge_base: str = ''
        with open(file) as f:
            for row in f:
                knowledge_base += row
        return tuprolog.theory.parsing.parse_theory(knowledge_base)





import unittest
import numpy as np
from tensorflow.python.ops.numpy_ops import argmax
from tensorflow import constant, float32, reshape, cast, stack, assert_equal, tile
from tensorflow.python.ops.array_ops import gather_nd
from psyki.logic.datalog.grammar.adapters.antlr4 import get_formula_from_string
from psyki.logic.datalog import Lukasiewicz
from test.resources.data import get_dataset
from test.resources.rules import get_rules
from test.resources.rules.poker import FEATURE_MAPPING as POKER_FEATURE_MAPPING, CLASS_MAPPING as POKER_CLASS_MAPPING


class TestLukasiewicz(unittest.TestCase):

    rules = list(get_rules('poker'))
    formulae = [get_formula_from_string(rule) for rule in rules]
    fuzzifier = Lukasiewicz(POKER_CLASS_MAPPING, POKER_FEATURE_MAPPING)
    functions = fuzzifier.visit(formulae)
    true = tile(reshape(constant(0.), [1, 1]), [1, 1])
    false = tile(reshape(constant(1.), [1, 1]), [1, 1])
    
    def test_nothing(self):
        hand1 = constant([2, 6, 2, 1, 4, 13, 2, 4, 4, 9], dtype=float32)
        hand2 = constant([4, 9, 3, 10, 4, 7, 4, 9, 3, 8], dtype=float32)
        output1 = constant([1, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=float32)
        output2 = constant([0, 1, 0, 0, 0, 0, 0, 0, 0, 0], dtype=float32)
        function = self.functions['nothing']

        # self._test_double_implication_hand_output_combinations(function, hand1, hand2, output1, output2)
        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def test_two(self):
        hand1 = constant([4, 9, 2, 2, 4, 2, 4, 6, 3, 9], dtype=float32)
        hand2 = constant([4, 1, 2, 2, 4, 7, 4, 10, 3, 9], dtype=float32)
        output1 = constant([0, 0, 1, 0, 0, 0, 0, 0, 0, 0], dtype=float32)
        output2 = constant([0, 0, 0, 1, 0, 0, 0, 0, 0, 0], dtype=float32)
        function = self.functions['two']

        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def test_flush(self):
        hand1 = constant([4, 4, 4, 13, 4, 7, 4, 11, 4, 1], dtype=float32)
        hand2 = constant([4, 4, 1, 13, 4, 7, 4, 11, 4, 1], dtype=float32)
        output1 = constant([0, 0, 0, 0, 0, 1, 0, 0, 0, 0], dtype=float32)
        output2 = constant([0, 0, 0, 0, 0, 0, 0, 1, 0, 0], dtype=float32)
        function = self.functions['flush']

        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def test_full(self):
        hand1 = constant([3, 2, 1, 2, 3, 11, 1, 11, 4, 11], dtype=float32)
        hand2 = constant([4, 1, 4, 2, 4, 7, 4, 10, 4, 9], dtype=float32)
        output1 = constant([0, 0, 0, 0, 0, 0, 1, 0, 0, 0], dtype=float32)
        output2 = constant([0, 0, 0, 1, 0, 0, 0, 0, 0, 0], dtype=float32)
        function = self.functions['full']

        # self._test_double_implication_hand_output_combinations(function, hand1, hand2, output1, output2)
        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def test_four(self):
        hand1 = constant([4, 9, 1, 9, 4, 7, 2, 9, 3, 9], dtype=float32)
        hand2 = constant([4, 9, 4, 5, 4, 7, 2, 9, 3, 9], dtype=float32)
        output1 = constant([0, 0, 0, 0, 0, 0, 0, 1, 0, 0], dtype=float32)
        output2 = constant([0, 0, 0, 1, 0, 0, 0, 0, 0, 0], dtype=float32)
        function = self.functions['four']

        # self._test_double_implication_hand_output_combinations(function, hand1, hand2, output1, output2)
        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def test_three(self):
        hand1 = constant([4, 9, 4, 2, 4, 7, 3, 9, 1, 9], dtype=float32)
        hand2 = constant([4, 1, 4, 2, 4, 7, 4, 10, 1, 9], dtype=float32)
        output1 = constant([0, 0, 0, 1, 0, 0, 0, 0, 0, 0], dtype=float32)
        output2 = constant([0, 0, 0, 1, 0, 0, 0, 1, 0, 0], dtype=float32)
        function = self.functions['three']

        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def test_pair(self):
        hand1 = constant([4, 9, 4, 2, 4, 7, 4, 6, 2, 9], dtype=float32)
        hand2 = constant([4, 1, 4, 2, 4, 7, 4, 10, 2, 9], dtype=float32)
        output1 = constant([0, 1, 0, 0, 0, 0, 0, 0, 0, 0], dtype=float32)
        output2 = constant([0, 0, 0, 1, 0, 0, 0, 0, 0, 0], dtype=float32)
        function = self.functions['pair']

        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def test_straight(self):
        hand1 = constant([1, 9, 4, 10, 2, 7, 4, 6, 3, 8], dtype=float32)
        hand2 = constant([1, 1, 4, 2, 2, 7, 4, 10, 3, 9], dtype=float32)
        output1 = constant([0, 0, 0, 0, 1, 0, 0, 0, 0, 0], dtype=float32)
        output2 = constant([0, 0, 0, 0, 0, 0, 1, 0, 0, 0], dtype=float32)
        function = self.functions['straight']

        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

        # Straight is also 10, 11, 12, 13, 1!
        hand3 = constant([1, 1, 4, 11, 2, 13, 4, 10, 3, 12], dtype=float32)
        hand3 = tile(reshape(hand3, [1, 10]), [1, 1])
        output1 = tile(reshape(output1, [1, 10]), [1, 1])
        result = reshape(function(hand3, output1), [1, 1])
        assert_equal(result, self.true)

    def test_straight_flush(self):
        hand1 = constant([4, 9, 4, 10, 4, 7, 4, 6, 4, 8], dtype=float32)
        hand2 = constant([4, 9, 3, 10, 4, 7, 4, 6, 3, 8], dtype=float32)
        output1 = constant([0, 0, 0, 0, 0, 0, 0, 0, 1, 0], dtype=float32)
        output2 = constant([0, 0, 0, 0, 0, 0, 1, 0, 0, 0], dtype=float32)
        function = self.functions['straight_flush']

        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def test_royal_flush(self):
        hand1 = constant([1, 10, 1, 11, 1, 13, 1, 12, 1, 1], dtype=float32)
        hand2 = constant([1, 9, 1, 11, 1, 13, 1, 10, 1, 12], dtype=float32)
        output1 = constant([0, 0, 0, 0, 0, 0, 0, 0, 0, 1], dtype=float32)
        output2 = constant([0, 0, 0, 0, 0, 0, 0, 0, 1, 0], dtype=float32)
        function = self.functions['royal_flush']

        # self._test_double_implication_hand_output_combinations(function, hand1, hand2, output1, output2)
        self._test_implication_hand_output_combinations(function, hand1, hand2, output1, output2)

    def _test_implication_hand_output_combinations(self, function, hand1, hand2, output1, output2) -> None:
        result1, result2, result3, result4 = self._get_combination_values(function, hand1, hand2, output1, output2)
        assert_equal(result1, self.true)
        assert_equal(result2, self.false)
        assert_equal(result3, self.true)
        assert_equal(result4, self.true)

    def _test_double_implication_hand_output_combinations(self, function, hand1, hand2, output1, output2) -> None:
        result1, result2, result3, result4 = self._get_combination_values(function, hand1, hand2, output1, output2)
        assert_equal(result1, self.true)
        assert_equal(result2, self.false)
        assert_equal(result3, self.false)
        assert_equal(result4, self.true)

    @staticmethod
    def _get_combination_values(function, hand1, hand2, output1, output2):
        hand1 = tile(reshape(hand1, [1, 10]), [1, 1])
        hand2 = tile(reshape(hand2, [1, 10]), [1, 1])
        output1 = tile(reshape(output1, [1, 10]), [1, 1])
        output2 = tile(reshape(output2, [1, 10]), [1, 1])
        result1 = reshape(function(hand1, output1), [1, 1])
        result2 = reshape(function(hand2, output1), [1, 1])
        result3 = reshape(function(hand1, output2), [1, 1])
        result4 = reshape(function(hand2, output2), [1, 1])
        return result1, result2, result3, result4

    def test_on_dataset(self):
        poker_training = get_dataset('poker', 'train')
        functions = [self.functions[name] for name, _ in sorted(POKER_CLASS_MAPPING.items(), key=lambda i: i[1])]
        train_x = poker_training[:, :-1]
        train_y = poker_training[:, -1]
        train_y = np.eye(10)[train_y.astype(int)]
        x, y = cast(train_x, dtype=float32), cast(train_y, dtype=float32)
        result = stack([reshape(function(x, y), [x.shape[0], 1]) for function in functions], axis=1)
        indices = stack([range(0, len(poker_training)), argmax(train_y, axis=1)], axis=1)
        assert_equal(gather_nd(result, indices), 0.)



from __future__ import annotations
from typing import Iterable, Callable, List
from tensorflow import Tensor, stack
from tensorflow.keras import Model
from tensorflow.keras.layers import Concatenate, Lambda
from tensorflow.keras.models import clone_model
from tensorflow.python.keras.saving.save import load_model
from psyki.logic.datalog.grammar import optimize_datalog_formula
from psyki.logic.datalog import Lukasiewicz, SubNetworkBuilder
from psyki.ski import Injector, Formula, Fuzzifier
from psyki.utils import eta, eta_one_abs, eta_abs_one


def _model_deep_copy(predictor: Model) -> Model:
    """
    Return a copy of the original model with the same weights.
    """
    new_predictor = clone_model(predictor)
    new_predictor.set_weights(predictor.get_weights())
    return new_predictor


class LambdaLayer(Injector):

    def __init__(self, predictor: Model, class_mapping: dict[str, int], feature_mapping: dict[str, int],
                 gamma: float = 1., fuzzifier: Fuzzifier = None):
        """
        @param predictor: the predictor.
        @param class_mapping: a map between constants representing the expected class in the logic formulae and the
        corresponding index for the predictor. Example:
            - 'setosa': 0,
            - 'virginica': 1,
            - 'versicolor': 2.
        @param feature_mapping: a map between variables in the logic formulae and indices of dataset features. Example:
            - 'PL': 0,
            - 'PW': 1,
            - 'SL': 2,
            - 'SW': 3.
        @param gamma: weight of the constraints.
        @param fuzzifier: the fuzzifier used to map the knowledge (by default it is Lukasiewicz).
        """
        self.predictor: Model = _model_deep_copy(predictor)
        self.class_mapping: dict[str, int] = class_mapping
        # self.feature_mapping: dict[str, int] = feature_mapping
        self.gamma: float = gamma
        # Use as default fuzzifier Lukasiewicz.
        self.fuzzifier = fuzzifier if fuzzifier is not None else Lukasiewicz(class_mapping, feature_mapping)
        self._fuzzy_functions: Iterable[Callable] = ()

    class ConstrainedModel(Model):

        def __init__(self, original_predictor: Model, constraints: Iterable[Callable], gamma: float = 1):
            self._gamma = gamma
            self._constraints = constraints
            self._input_shape = original_predictor.input_shape
            predictor_output = original_predictor.layers[-1].output
            x = Concatenate(axis=1)([original_predictor.input, predictor_output])
            x = Lambda(self._cost, original_predictor.output.shape)(x)
            super().__init__(original_predictor.inputs, x)

        def call(self, inputs, training=None, mask=None):
            return super().call(inputs, training, mask)

        def get_config(self):
            pass

        def set_gamma(self, gamma: float = 1):
            self._gamma = gamma

        def remove_constraints(self) -> Model:
            """
            Remove the lambda layer obtained by the injected rules.
            """
            # Layer -3 is the layer before the lambda layer (last original layer -> lambda -> output).
            return Model(self.input, self.layers[-3].output)

        def _cost(self, output_layer: Tensor) -> Tensor:
            input_len = self._input_shape[1]
            x, y = output_layer[:, :input_len], output_layer[:, input_len:]
            cost = stack([function(x, y) for function in self._constraints], axis=1)
            return y + (cost * self._gamma)

    def inject(self, rules: List[Formula]) -> Model:
        dict_functions = self.fuzzifier.visit(rules)
        # To ensure that every function refers to the right class we check the associated class name.
        self._fuzzy_functions = [dict_functions[name] for name, _ in
                                 sorted(self.class_mapping.items(), key=lambda i: i[1])]
        return self.ConstrainedModel(_model_deep_copy(self.predictor), self._fuzzy_functions, self.gamma)

    def _clear(self):
        self._fuzzy_functions = ()

    def load(self, file):
        """
        Use this function to load a trained model.
        """
        # return load_model(file, custom_objects={'_cost': self._cost})
        raise Exception("Load of constrained model not implemented yet")


class NetworkComposer(Injector):
    """
    This injector builds a set of moduls, aka ad-hoc layers, and inserts them into the predictor (a neural network).
    In this way the predictor can exploit the knowledge via these modules which mimic the logic formulae.
    With the default fuzzifier this is the implementation of KINS: Knowledge injection via network structuring.
    """

    def __init__(self, predictor: Model, feature_mapping: dict[str, int], layer: int = 0, fuzzifier: Fuzzifier = None):
        """
        @param predictor: the predictor.
        @param feature_mapping: a map between variables in the logic formulae and indices of dataset features. Example:
            - 'PL': 0,
            - 'PW': 1,
            - 'SL': 2,
            - 'SW': 3.
        @param layer: the level of the layer where to perform the injection.
        @param fuzzifier: the fuzzifier used to map the knowledge (by default it is SubNetworkBuilder).
        """
        self.predictor: Model = _model_deep_copy(predictor)
        # self.feature_mapping: dict[str, int] = feature_mapping
        if layer < 0 or layer > len(predictor.layers) - 2:
            raise Exception('Cannot inject knowledge into layer ' + str(layer) +
                            '.\nYou can inject from layer 0 to ' + str(len(predictor.layers) - 2))
        self.layer = layer
        # Use as default fuzzifier SubNetworkBuilder.
        self.fuzzifier = fuzzifier if fuzzifier is not None else SubNetworkBuilder(self.predictor.input, feature_mapping)
        self._fuzzy_functions: Iterable[Callable] = ()

    def inject(self, rules: List[Formula]) -> Model:
        self._clear()
        # Prevent side effect on the original rules during optimization.
        rules_copy = [rule.copy() for rule in rules]
        for rule in rules_copy:
            optimize_datalog_formula(rule)
        predictor_input: Tensor = self.predictor.input
        modules = self.fuzzifier.visit(rules_copy)
        if self.layer == 0:
            # Injection!
            x = Concatenate(axis=1)([predictor_input] + modules)
            self.predictor.layers[1].build(x.shape)
            for layer in self.predictor.layers[1:]:
                x = layer(x)
        else:
            x = self.predictor.layers[1](predictor_input)
            for layer in self.predictor.layers[2:self.layer + 1]:
                x = layer(x)
            # Injection!
            x = Concatenate(axis=1)([x] + modules)
            self.predictor.layers[self.layer + 1].build(x.shape)
            x = self.predictor.layers[self.layer + 1](x)
            for layer in self.predictor.layers[self.layer + 2:]:
                # Correct shape if needed (e.g., dropout layers)
                if layer.input_shape != x.shape:
                    layer.build(x.shape)
                x = layer(x)
        new_predictor = Model(predictor_input, x)
        # TODO: clone all old weights into the same layers

        return new_predictor

    @staticmethod
    def load(file: str):
        return load_model(file, custom_objects={'eta': eta, 'eta_one_abs': eta_one_abs, 'eta_abs_one': eta_abs_one})

    def _clear(self):
        self._fuzzy_functions = ()



from tuprolog.core.parsing import parse_struct
from psyki.ski import Formula


class PrologFormula(Formula):

    def __init__(self, formula: str):
        self.string: str = formula
        self.theory = parse_struct(formula)



import re
from test.resources.data.splice_junction import AGGREGATE_FEATURE_MAPPING

VARIABLE_BASE_NAME = 'X'
AND_SYMBOL = ' ∧ '
OR_SYMBOL = ' ∨ '
NOT_SYMBOL = '¬'
LESS_EQUAL_SYMBOL = ' ≤ '
PLUS_SYMBOL = ' + '
STATIC_IMPLICATION_SYMBOL = ' ← '
MUTABLE_IMPLICATION_SYMBOL = ' ⇐ '
STATIC_RULE_SYMBOL = '::-'
MUTABLE_RULE_SYMBOL = ':-'
RULE_DEFINITION_SYMBOLS = (STATIC_RULE_SYMBOL, MUTABLE_RULE_SYMBOL)
RULE_DEFINITION_SYMBOLS_REGEX = '(' + '|'.join(RULE_DEFINITION_SYMBOLS) + ')'


def next_index(index: str, indices: list[int], offset: int) -> int:
    new_index: int = int(index) + offset
    modified: bool = False
    while new_index not in indices:
        new_index += 1
        modified = True
    return new_index + previous_holes(indices, indices.index(new_index)) if not modified else new_index


def previous_holes(l: list[int], i: int) -> int:
    j = 0
    for k in list(range(0, i)):
        if l[k] + 1 != l[k + 1]:
            j += 1
    return j


def explicit_variables(e: str) -> str:
    result = ''
    for key in AGGREGATE_FEATURE_MAPPING.keys():
        if key.lower() in e:
            values = [v for v in AGGREGATE_FEATURE_MAPPING[key]]
            if len(values) > 1:
                result += AND_SYMBOL.join(
                    NOT_SYMBOL + '(' + re.sub(key.lower(), value.lower(), e) + ')' for value in values)
    return NOT_SYMBOL + '(' + result + ')' if result != '' else e


def replace(s: str, e: str) -> str:
    values = [v for v in AGGREGATE_FEATURE_MAPPING[s]]
    return '(' + OR_SYMBOL.join('(' + re.sub(s.lower(), value.lower(), e) + ')' for value in values) + ')'



import unittest
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from tensorflow.keras import Input, Model
from tensorflow.python.framework.random_seed import set_random_seed
from psyki.logic.datalog.grammar.adapters.antlr4 import get_formula_from_string
from psyki.ski.injectors import LambdaLayer, NetworkComposer
from test.resources.rules import get_rules
from test.utils import get_mlp


EPOCHS = 50
BATCH_SIZE = 8
VERBOSE = 0
x, y = load_iris(return_X_y=True, as_frame=True)
encoder = OneHotEncoder(sparse=False)
encoder.fit_transform([y])
dataset = x.join(y)
train, test = train_test_split(dataset, test_size=0.5, random_state=0)
train_x, train_y = train.iloc[:, :-1], train.iloc[:, -1]
test_x, test_y = test.iloc[:, :-1], test.iloc[:, -1]
class_mapping = {'setosa': 0, 'virginica': 1, 'versicolor': 2}
variable_mapping = {'PL': 0, 'PW': 1, 'SL': 2, 'SW': 3}


class TestInjection(unittest.TestCase):

    def test_lambda_layer_on_iris(self):
        set_random_seed(0)
        formulae = [get_formula_from_string(rule) for rule in get_rules('iris')]
        input_layer = Input((4,))
        predictor = get_mlp(input_layer, 3, 3, 32, 'relu', 'softmax')
        predictor = Model(input_layer, predictor)
        injector = LambdaLayer(predictor, class_mapping, variable_mapping, gamma=0.5)
        model = injector.inject(formulae)

        compile_and_train(model)
        model = model.remove_constraints()
        model.compile('adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'])
        accuracy = model.evaluate(test_x, test_y)[1]
        self.assertTrue(accuracy > 0.986)

    def test_network_composer_on_iris(self):
        set_random_seed(0)
        formulae = [get_formula_from_string(rule) for rule in get_rules('iris')]
        input_layer = Input((4,))
        predictor = get_mlp(input_layer, 3, 3, 32, 'relu', 'softmax')
        predictor = Model(input_layer, predictor)
        injector = NetworkComposer(predictor, variable_mapping, 2)
        model = injector.inject(formulae)

        compile_and_train(model)
        accuracy = model.evaluate(test_x, test_y)[1]
        self.assertTrue(accuracy > 0.973)


def compile_and_train(model):
    model.compile('adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'])
    model.fit(train_x, train_y, batch_size=BATCH_SIZE, epochs=EPOCHS, verbose=VERBOSE)



from __future__ import annotations
from abc import ABC
from typing import Iterable
from psyki.ski import Formula
import psyki.logic.datalog as datalog


# TODO: refactoring
def optimize_datalog_formula(formula: Formula):
    if isinstance(formula, datalog.grammar.Expression):
        lhs = formula.lhs
        rhs = formula.rhs
        op = formula.op
        if op in ('∧', '∨', '+') and len(formula.nary) == 0:
            if isinstance(lhs, datalog.grammar.Expression):
                if lhs.op == op:
                    optimize_datalog_formula(lhs)
                    for clause in lhs.nary:
                        formula.nary.append(clause)
                    formula.lhs = None
                else:
                    formula.nary.append(lhs)
            else:
                formula.nary.append(lhs)
            if isinstance(rhs, datalog.grammar.Expression):
                if rhs.op == op:
                    optimize_datalog_formula(rhs)
                    for clause in rhs.nary:
                        formula.nary.append(clause)
                    formula.rhs = None
                else:
                    formula.nary.append(rhs)
            else:
                formula.nary.append(rhs)
        else:
            optimize_datalog_formula(lhs)
            optimize_datalog_formula(rhs)
    else:
        if hasattr(formula, 'lhs'):
            optimize_datalog_formula(formula.lhs)
        if hasattr(formula, 'rhs'):
            optimize_datalog_formula(formula.rhs)
        if hasattr(formula, 'predicate'):
            optimize_datalog_formula(formula.predicate)


class DatalogFormula(Formula):

    def __init__(self, lhs: DefinitionClause, rhs: Clause, op: str):
        self.lhs: DefinitionClause = lhs
        self.rhs: Clause = rhs
        self.op: str = op

    def __str__(self) -> str:
        return str(self.lhs) + self.op + str(self.rhs)

    def copy(self) -> Formula:
        return DatalogFormula(self.lhs.copy(), self.rhs.copy(), self.op)


class DefinitionClause(Formula):

    def __init__(self, predication: str, arg: Argument):
        self.predication: str = predication
        self.arg: Argument = arg

    def __str__(self) -> str:
        return self.predication + '(' + str(self.arg) + ')'

    def copy(self) -> Formula:
        return DefinitionClause(self.predication, self.arg)


class Clause(Formula, ABC):
    pass


class Expression(Clause):

    def __init__(self, lhs: Clause, rhs: Clause, op: str, nary: Iterable[Clause] = []):
        self.lhs: Clause = lhs
        self.rhs: Clause = rhs
        self.nary: list[Clause] = list(nary)
        self.op: str = op

    def __str__(self) -> str:
        if len(self.nary) == 0:
            return '((' + str(self.lhs) + ')' + self.op + '(' + str(self.rhs) + '))'
        else:
            return '(' + self.op + '(' + ','.join(str(clause) for clause in self.nary) + ')'

    def copy(self) -> Formula:
        return Expression(self.lhs.copy(), self.rhs.copy(), self.op, [c.copy() for c in self.nary])


class Literal(Clause, ABC):
    pass


class Negation(Literal):

    def __init__(self, predicate: Clause):
        self.predicate: Clause = predicate

    def __str__(self) -> str:
        return 'neg(' + str(self.predicate) + ')'

    def copy(self) -> Formula:
        return Negation(self.predicate.copy())


class Predicate(Clause, ABC):
    pass


class Unary(Predicate):

    def __init__(self, name: str):
        self.name: str = name

    def __str__(self) -> str:
        return self.name

    def copy(self) -> Formula:
        return Unary(self.name)


class Nary(Predicate):

    def __init__(self, name: str, arg: Argument):
        self.name: str = name
        self.arg: Argument = arg

    def __str__(self) -> str:
        return self.name + '(' + str(self.arg) + ')'

    def copy(self) -> Formula:
        return Nary(self.name, self.arg)


class Term(Predicate, ABC):
    pass


class Constant(Term, ABC):
    pass


class Predication(Constant):

    def __init__(self, name: str):
        self.name: str = name

    def __str__(self) -> str:
        return self.name

    def copy(self) -> Formula:
        return Predication(self.name)


class Number(Constant):

    def __init__(self, value: str):
        self.value: float = float(value)

    def __str__(self) -> str:
        return str(self.value)

    def copy(self) -> Formula:
        return Number(self.value)


class Variable(Term):

    def __init__(self, name: str):
        self.name: str = name

    def __str__(self) -> str:
        return self.name

    def copy(self) -> Formula:
        return Variable(self.name)


class Argument(Formula):

    def __init__(self, term: Term, arg: Argument = None):
        self.term: Term = term
        self.arg: Argument = arg

    def __str__(self) -> str:
        return str(self.term) + (',' + str(self.arg) if self.arg is not None else '')

    def copy(self) -> Formula:
        return Argument(self.term.copy(), self.arg)



import re
from os import popen, system
from pathlib import Path
from psyki import PSYKI_PATH

PATH = Path(__file__).parents[0]


def create_antlr4_parser(file: str, path: str):
    antlr4_version = re.split(r'=', popen('cat ' + str(PSYKI_PATH / 'requirements.txt') + ' | grep antlr4').read())
    if hasattr(antlr4_version, '__len__') and len(antlr4_version) > 1:
        antlr4_version = antlr4_version[1]
        if hasattr(antlr4_version, '__len__') and len(antlr4_version) > 0:
            antlr4_version = antlr4_version[:-1]
    else:
        antlr4_version = '4.9.3'
    system('wget https://www.antlr.org/download/antlr-' + antlr4_version + '-complete.jar -O' + str(PSYKI_PATH / ('antlr-' + antlr4_version + '-complete.jar')))
    system('export CLASSPATH="' + str(PSYKI_PATH / ('antlr-' + antlr4_version + '-complete.jar:$CLASSPATH"')))
    system('java -jar ' + str(PSYKI_PATH / ('antlr-' + antlr4_version + '-complete.jar -Dlanguage=Python3 ' + file + ' -visitor -o ' + path)))
    system('rm ' + str(PSYKI_PATH / ('antlr-' + antlr4_version + '-complete.jar')))



from __future__ import annotations
from tensorflow import minimum, maximum, abs
from tensorflow.python.types.core import Tensor


def eta(x: Tensor) -> Tensor:
    return minimum(1., maximum(0., x))


def eta_abs(x: Tensor) -> Tensor:
    return eta(abs(x))


def eta_abs_one(x: Tensor) -> Tensor:
    return eta(abs(x - 1.))


def eta_one_abs(x: Tensor) -> Tensor:
    return eta(1. - abs(x))



from pathlib import Path
from typing import Iterable, Callable
import numpy as np
from test.resources.data.splice_junction import get_indices
from test.resources.rules.utils import *

PATH = Path(__file__).parents[0]
INDEX_IDENTIFIER = '@'
NOT_IDENTIFIER = 'not'
CONFUSION_MATRIX = np.array([[295, 0, 473], [25, 31, 711], [3, 0, 1652]])
CLASS_LABELS = ['IE', 'EI', 'N']


def parse_clause(rest: str, rhs: str = '', aggregation: str = AND_SYMBOL) -> str:
    for j, clause in enumerate(rest.split(',')):
        index = re.match(INDEX_IDENTIFIER + '[-]?[0-9]*', clause)
        negation = re.match(NOT_IDENTIFIER, clause)
        n = re.match('[0-9]*of', clause)
        if index is not None:
            index = clause[index.regs[0][0]:index.regs[0][1]]
            clause = clause[len(index):]
            clause = re.sub('\'', '', clause)
            index = index[1:]
            rhs += aggregation.join(explicit_variables(
                VARIABLE_BASE_NAME + ('_' if next_index(index, get_indices(), i) < 0 else '') +
                str(abs(next_index(index, get_indices(), i))) +
                ' = ' + value.lower()) for i, value in enumerate(clause))
        elif negation is not None:
            new_clause = re.sub(NOT_IDENTIFIER, NOT_SYMBOL, clause)
            new_clause = re.sub('-', '_', new_clause.lower())
            new_clause = re.sub('\)', '())', new_clause)
            rhs += new_clause
        elif n is not None:
            new_clause = clause[n.regs[0][1]:]
            new_clause = re.sub('\(|\)', '', new_clause)
            inner_clause = parse_clause(new_clause, rhs, PLUS_SYMBOL)
            inner_clause = '(' + (')' + PLUS_SYMBOL + '(').join(e for e in inner_clause.split(PLUS_SYMBOL)) + ')'
            n = clause[n.regs[0][0]:n.regs[0][1] - 2]
            rhs += n + LESS_EQUAL_SYMBOL + '(' + inner_clause + ')'
        else:
            rhs += re.sub('-', '_', clause.lower()) + '()'
        if j < len(rest.split(',')) - 1:
            rhs += AND_SYMBOL
    return rhs


def get_binary_datalog_rules(rules: Iterable[str]) -> Iterable[str]:
    results = []
    term_regex = '[a-z]+'
    variable_regex = VARIABLE_BASE_NAME + '[_]?[0-9]+'
    regex = variable_regex + '[ ]?=[ ]?' + term_regex
    for rule in rules:
        tmp_rule = rule
        partial_result = ''
        while re.search(regex, tmp_rule) is not None:
            match = re.search(regex, tmp_rule)
            start, end = match.regs[0]
            matched_string = tmp_rule[start:end]
            ante = tmp_rule[:start]
            medio = matched_string[:re.search(variable_regex, matched_string).regs[0][1]] + \
                    matched_string[re.search(term_regex, matched_string).regs[0][0]:] + ' = 1'
            partial_result += ante + medio
            tmp_rule = tmp_rule[end:]
        partial_result += tmp_rule
        results.append(partial_result)
    return results


def get_splice_junction_datalog_rules(rules: Iterable[str]) -> Iterable[str]:
    return get_datalog_rules(rules, {'ei', 'ie', 'n'}, parse_clause)


def get_datalog_rules(rules: Iterable[str], class_labels: set[str], parse_clause_f: Callable) -> Iterable[str]:
    results = []
    for rule in rules:
        rule = re.sub(r' |\.', '', rule)
        name, op, rest = re.split(RULE_DEFINITION_SYMBOLS_REGEX, rule)
        name = re.sub('-', '_', name.lower())
        rhs = parse_clause_f(rest)
        if name in class_labels:
            results.append('class(' + name + ')' +
                           (STATIC_IMPLICATION_SYMBOL if op == STATIC_RULE_SYMBOL else MUTABLE_IMPLICATION_SYMBOL)
                           + rhs)
        results.append(name + '(' + ')' +
                       (STATIC_IMPLICATION_SYMBOL if op == STATIC_RULE_SYMBOL else MUTABLE_IMPLICATION_SYMBOL) + rhs)
    return results

